#include "lattice.h"
#include "stopwatch.h"
#include <iomanip>
#include <fstream>
#include <sstream>
#include "filescanner.h"
#include "paramfile.h"

Gascoigne::Timer LatticeTimer;

namespace GasParticle
{
  extern int NPARTICLES;

  ////////////////////////////////////////////////// Init

  void Lattice::BasicInit(const Gascoigne::ParamFile &_paramfile)
  {
    NPARTICLES = 0;
  }

  void Lattice::InitBoundaries()
  {

    // allocate boundary lines to elements. !!!! Not Efficient at all!
    // check all element vs. all boundaries!
    int nnn = 0;
    for (auto &el : elements)
      {
	// we add all boundaries that have distance H from the element mid-point
	// might be too much, but otherwise we might miss boundaries that are
	// very close to the element edge in the next element
	// !!! distance H ist not enough. We need H + Particle-Radius
	double partradius = 3.e-3;

	// midpoint of element
	double xm = 0.5 * (el.X0 + el.X1);
	double ym = 0.5 * (el.Y0 + el.Y1);

	for (int bli = 0; bli < boundary_lines.size(); ++bli)
	  {
	    double dist = boundary_lines[bli].raw_distance(xm, ym);
	    if (fabs(dist) < el.H + partradius)
	      {
		el.boundary_lines_index.push_back(bli);
		++nnn;
	      }
	  }

	for (int bci = 0; bci < boundary_circles.size(); ++bci)
	  {
	    double dist = boundary_circles[bci].raw_distance(xm, ym);
	    if (fabs(dist) < el.H + partradius)
	      {
		el.boundary_circles_index.push_back(bci);
		++nnn;
	      }
	  }
      }
    std::cout << std::endl
	      << "Created Lattice" << std::endl
	      << "    #boundarylines  " << boundary_lines.size() << " / " << nnn << std::endl
	      << "    #boundarycircles " << boundary_circles.size() << " / " << nnn << std::endl;
  }

  void Lattice::InitFromGascoigneMesh(const Gascoigne::GascoigneMesh2d *M, double _H)
  {
    LatticeTimer.start("InitFromGascoigneMesh\t\t");

    assert(M);
    H = _H;
    assert(H > 0);

    ////////// Get data from mesh
    const std::vector<Gascoigne::Vertex<2>> &Mvertices = M->GetVertexVector();

    ////////// Global Lattice data
    LatticeTimer.start("InitFromGascoigneMesh::Structure");
    // bounding box
    X0 = 1.e10;
    Y0 = 1.e10;
    X1 = -1.e10;
    Y1 = -1.e10;
    for (auto it : Mvertices)
      {
	if (it.x() < X0)
	  X0 = it.x();
	if (it.y() < Y0)
	  Y0 = it.y();
	if (it.x() > X1)
	  X1 = it.x();
	if (it.y() > Y1)
	  Y1 = it.y();
      }
    assert(X1 - X0 > 0);
    assert(Y1 - Y0 > 0);

    // enlarge lattice to avoid boundary trouble
    // X0 -= 0.01 * H;
    // X1 += 0.01 * H;
    // Y0 -= 0.01 * H;
    // Y1 += 0.01 * H;
    // computing no of elements
    NX = ceil((X1 - X0) / H);
    NY = ceil((Y1 - Y0) / H);

    // adjust and compute size  of lattice
    X1 = X0 + H * NX;
    Y1 = Y0 + H * NY;

    //create inp file with lattice
    std::remove("inplattice.inp");
    std::ofstream inpfile("inplattice.inp", std::ios::app);
    //header
    inpfile << (NX + 1) * (NY + 1) << " " << NX * NY + 2 * (NX + NY) << " 0 0 0" << std::endl;
    //nodes
    int non = 0;
    for (int i = 0; i < NY + 1; i++)
      for (int j = 0; j < NX + 1; j++)
	{
	  inpfile << non << " " << X0 + j * H << " " << Y0 + i * H << " 0" << std::endl;
	  ++non;
	}
    //quads
    int noq = 0;
    for (int i = 0; i < NY; i++)
      for (int j = 0; j < NX; j++)
	{
	  inpfile << noq << " 3 quad " << i * (NX + 1) + j << " " << i * (NX + 1) + j + 1 << " " << (i + 1) * (NX + 1) + j + 1 << " " << (i + 1) * (NX + 1) + j << std::endl;
	  ++noq;
	}
    inpfile.close();

    // create lattice structure
    elements.resize(NX * NY);
    for (int iy = 0; iy < NY; ++iy)
      for (int ix = 0; ix < NX; ++ix)
	element(ix, iy).BasicInit(X0 + ix * H, Y0 + iy * H, H);
    LatticeTimer.stop("InitFromGascoigneMesh::Structure");
    ////////////////////////////// create boundary structure // NOT FROM GASCOIGNE MESH
    boundary_lines.clear();
    boundary_circles.clear();

    std::cout << std::endl
	      << "Created Lattice" << std::endl
	      << "    H               " << H << std::endl
	      << "    [X0,X1]x[Y0,Y1] "
	      << "[" << X0 << " , " << X1 << " ] x [ " << Y0 << " , " << Y1 << " ]" << std::endl
	      << "    NX/NY           " << NX << " / " << NY << std::endl
	      << "    #elements       " << NX * NY << std::endl
	      << "    #vertices       " << (NX + 1) * (NY + 1) << std::endl;
    LatticeTimer.stop("InitFromGascoigneMesh\t\t");

  }

  // remove all particles with marked for deletion (size<0)
  void Lattice::clear_deleted_particles()
  {
#pragma omp parallel for
    for (int e = 0; e < elements.size(); ++e)
      elements[e].clear_deleted_particles();

    for (int e = 0; e < elements.size(); ++e)
      for (int p = 0; p < elements[e]._P.size(); ++p)
	assert(elements[e]._P[p].radius > 0);
  }

  // delete all particles outside the domain
  void Lattice::delete_outside_particles()
  {
#pragma omp parallel for
    for (int e = 0; e < elements.size(); ++e)
      elements[e].delete_outside_particles(boundary_lines, boundary_circles);
  }

  // delete all particles that overlap with each other
  void Lattice::delete_overlapping_particles()
  {
    // we do not delete diagonal overlaps!

#pragma omp parallel for
    for (int iy = 0; iy < NY; ++iy)
      for (int ix = 0; ix < NX; ++ix)
	{
	  int eid = e_id(ix, iy);
	  elements[eid].delete_overlapping_particles(); // inside one element
	  if (ix + 1 < NX)
	    elements[eid].delete_overlapping_particles(elements[eid + 1]); // to the right
	}

#pragma omp parallel for
    for (int ix = 0; ix < NX; ++ix)
      for (int iy = 0; iy < NY; ++iy)
	{
	  int eid = e_id(ix, iy);
	  if (iy + 1 < NY)
	    elements[eid].delete_overlapping_particles(elements[eid + NX]); // to the top
	}
  }

  //////////// I/O
  void Lattice::write_vtk(const std::string &name, int i)
  {
    clear_deleted_particles();

    LatticeTimer.start("write_vtk\t\t\t");
    char s[50];
    sprintf(s, "%s.%05i.vtk", name.c_str(), i);
    std::ofstream VTK;
    VTK.open(s);

    int np = nparticles();

    VTK << "# vtk DataFile Version 2.0" << std::endl
	<< "Unstructured Grid Example" << std::endl
	<< "ASCII" << std::endl
	<< std::endl
	<< "DATASET UNSTRUCTURED_GRID" << std::endl
	<< "POINTS " << np << " float" << std::endl;
    for (const auto &el : elements)
      for (const auto &p : el._P)
	VTK << p.x << " " << p.y << " 0" << std::endl;

    // output size
    VTK << std::endl
	<< "POINT_DATA " << np << std::endl
	<< std::endl
	<< "SCALARS size float 1" << std::endl
	<< "LOOKUP_TABLE default" << std::endl;
    for (const auto &el : elements)
      for (const auto &p : el._P)
	VTK << p.radius << " ";
    VTK << std::endl;

    VTK << "SCALARS height float 1" << std::endl
	<< "LOOKUP_TABLE default" << std::endl;
    for (const auto &el : elements)
      for (const auto &p : el._P)
	VTK << p.height << " ";
    VTK << std::endl;

    VTK.close();

    LatticeTimer.stop("write_vtk\t\t\t");
  }

  //////////// I/O
  void Lattice::write_lattice_vtk(const std::string &name, int i)
  {
    clear_deleted_particles();

    LatticeTimer.start("write_vtk\t\t\t");
    char s[50];
    sprintf(s, "%s.%05i.vtk", name.c_str(), i);
    std::ofstream VTK;
    VTK.open(s);

    VTK << "# vtk DataFile Version 2.0" << std::endl
	<< "Unstructured Grid Example" << std::endl
	<< "ASCII" << std::endl
	<< std::endl
	<< "DATASET UNSTRUCTURED_GRID" << std::endl
	<< "POINTS " << (NX + 1) * (NY + 1) << " float" << std::endl;
    for (int iy = 0; iy <= NY; ++iy)
      for (int ix = 0; ix <= NX; ++ix)
	VTK << X0 + ix * H << " " << Y0 + iy * H << " 0" << std::endl;
    VTK << std::endl
	<< "POINT_DATA " << (NX + 1) * (NY + 1) << std::endl
	<< std::endl
	<< "SCALARS size float 1" << std::endl
	<< "LOOKUP_TABLE default" << std::endl;
    for (int iy = 0; iy <= NY; ++iy)
      for (int ix = 0; ix <= NX; ++ix)
	VTK << VX[(1 + NX) * iy + ix] << " ";
    VTK << std::endl;

    VTK.close();

    LatticeTimer.stop("write_vtk\t\t\t");
  }

}
