/*----------------------------   boundaryline.h ---------------------------*/
/*      $Id:$                 */
#ifndef __boundaryline_H
#define __boundaryline_H
/*----------------------------   boundaryline.h ---------------------------*/

#include <cassert>
#include <cmath>
#include "particle.h"

namespace GasParticle
{

  class BoundaryLine
  {
  public:
    double x0, y0, x1, y1; // coordinates of the two points
    double nx, ny;         // normal vector pointing out
    double tx, ty;         // tangential vector pointing from x0 to x1
    int color;

  public:
    // Init
    BoundaryLine(double _x0, double _y0, double _x1, double _y1, int _color)
        : x0(_x0), y0(_y0), x1(_x1), y1(_y1), color(_color)
    {
      nx = (y1 - y0); // points outward
      ny = -(x1 - x0);
      std::cout << "(" << nx <<  "," << ny <<  ")  " << std::endl;
      double norm = nx * nx + ny * ny;
      assert(norm > 0);
      nx /= sqrt(norm);
      ny /= sqrt(norm);
      tx = (x1 - x0); // points outward
      ty = (y1 - y0);
      norm = tx * tx + ty * ty;
      assert(norm > 0);
      tx /= sqrt(norm);
      ty /= sqrt(norm);
    }
    bool operator==(const BoundaryLine &B) const
    {
      return ((x0 == B.x0) && (x1 == B.x1) && (y0 == B.y0) &&
              (y1 == B.y1) && (color == B.color));
    }

    ///////////////////////// init
    // get distance to a point
    double raw_distance(double x, double y) const
    {
      return nx * (x - x0) + ny * (y - y0);
    }

    //////////////////// collision
    // get signed distance to wall. positive: outside
    double distance(const Particle &p) const
    {
      // beyond the end of the line
      if (tx * (p.x - x1) + ty * (p.y - y1) > 0)
        return -1;
      // before the beginning of the line
      if (tx * (x0 - p.x) + ty * (y0 - p.y) > 0)
        return -1;

      // Distance to the particle center! particles can overlap with the boundary
      return nx * (p.x - x0) + ny * (p.y - y0);
      // return p.radius + nx * (p.x - x0) + ny * (p.y - y0);
    }
    // flip particle at wall
    void flip(double d, Particle &p) const
    {
      assert(d > 0); // assume positive distance, outside

      // compute time of first contact
  

      double nv = nx * p.vx + ny * p.vy;
      assert(nv > 0);

      double t = -d / nv;
      // assert(t<0);
      // TODO set the way it makes sense (depending on the time step)
      // assert(t>-1.e-4);

      // move particle back
      p.x += t * p.vx;
      p.y += t * p.vy;
  
      // flip velocity
      p.vx -= (1+0.9) * nv * nx; // -( 1+alpha) 
      p.vy -= (1+0.9) * nv * ny;

      // move particle forward with new velocity
      p.x -= t * p.vx;
      p.y -= t * p.vy;
      
    }

    // return s such that xi + s (xo-xi) is exactly on the boundary
    double position_between_points(double xi, double yi, double xo, double yo) const
    {
      assert(raw_distance(xi, yi) * raw_distance(xo, yo) < 0);
      double s = ((x0 - xi) * nx + (y0 - yi) * ny) / ((xo - xi) * nx + (yo - yi) * ny);
      assert(s > 0);
      assert(s < 1);
      return s;
    }
  };

  class BoundaryCircle
  {
  public:
    double x, y, r; // midpoint, radius
    int color;

  public:
    // Init
    BoundaryCircle(double _x, double _y, double _r, int _color)
        : x(_x), y(_y), r(_r), color(_color)
    {
    }
    bool operator==(const BoundaryCircle &B) const
    {
      return ((x == B.x) && (r == B.r) &&
              (y == B.y) && (color == B.color));
    }

    ///////////////////////// init
    // get distance to a point
    double raw_distance(double xp, double yp) const
    {
      return r - sqrt((x - xp) * (x - xp) + (y - yp) * (y - yp));
    }

    //////////////////// collision
    // get signed distance to wall. positive: outside
    double distance(const Particle &p) const
    {
      assert(p.radius > 0);
      assert(r > 0);
      // distance from particle midpoint to ball midpoint
      double dm = sqrt((x - p.x) * (x - p.x) + (y - p.y) * (y - p.y));
      // distance from outer surfaces. positive overlap and outside
      return r + p.radius - dm;
    }
    // flip particle at wall
    void flip(double d, Particle &p) const
    {
      assert(d > 0); // outer distance must be positive, so overlap or outside domain
      // compute first contact time (in past)
      double vv = p.vnorm2();                                 // velocity squared
      double p2 = (p.vx * (p.x - x) + p.vy * (p.y - y)) / vv; // p/2
      double q = ((p.x - x) * (p.x - x) + (p.y - y) * (p.y - y) - (p.radius + r) * (p.radius + r)) / vv;
      double Disc = p2 * p2 - q;
      assert(Disc > 0);
      double t = -(p2 + sqrt(Disc));
      

      // move back till first contact
      p.x += t * p.vx;
      p.y += t * p.vy;

      // compute contact normal
      double nx = x - p.x;
      double ny = y - p.y;
      double nn = sqrt(nx * nx + ny * ny);
      nx /= nn;
      ny /= nn;

      // flip velocities
      double vd = p.vx * nx + p.vy * ny;
      p.vx -= 2 * vd * nx;
      p.vy -= 2 * vd * ny;

      // move forward after contact
      p.x -= t * p.vx;
      p.y -= t * p.vy;
    }

    // return s such that xi + s (xo-xi) is exactly on the boundary
    double position_between_points(double xi, double yi, double xo, double yo) const
    {
      assert(raw_distance(xi, yi) * raw_distance(xo, yo) < 0);
      double delta = 4. * (pow(xo - xi, 2) + pow(yo - yi, 2)) * pow(r, 2) - pow((xo - xi) * (yi - y) - (yo - yi) * (xi - x), 2);
      double sqrtdelta = sqrt(delta);
      double s1 = (-2. * ((xo - xi) * (xi - x) + (yo - yi) * (yi - y)) - sqrtdelta) / (2. * (pow(xo - xi, 2) + pow(yo - yi, 2)));
      double s2 = (-2. * ((xo - xi) * (xi - x) + (yo - yi) * (yi - y)) + sqrtdelta) / (2. * (pow(xo - xi, 2) + pow(yo - yi, 2)));
      if (s1 >= 0 || s1 <= 1)
        return s1;
      if (s2 >= 0 || s2 <= 1)
        return s2;
    }
  };

} // namespace GasParticle

/*----------------------------   boundaryline.h ---------------------------*/
/* end of #ifndef __boundaryline_H */
#endif
/*----------------------------   boundaryline.h ---------------------------*/
