/*----------------------------   lattice.h     ---------------------------*/
/*      $Id:$                 */
#ifndef __lattice_H
#define __lattice_H
/*----------------------------   lattice.h     ---------------------------*/

#include "gascoignemesh2d.h"
#include "element.h"
#include <string>
#include <random> // Funktionen fuer Zufallszahlen
#include "compvector.h"


#include "stopwatch.h"
extern double TIME;

extern Gascoigne::Timer CoupledTimer;

namespace GasParticle
{
  /**
   * The lattice is a regular grid of _NX * NY elements of size H*H
   * - The lower left point has the coordinate (_X0,_Y0)
   * - The upper right point has the coordinate (_X1,_Y1)
   * The elements, stored in the vector _E contain the particles and manage boundaries
   *
   *
   * ELEMENT ID:
   * - each element has an element id   0 <= e_id < _NX*NY
   * - and two element coordinates e_id = e_iy * NX + e_ix
   *
   * NODE ID:
   * - the lattice has nodes with id   0 <= n_id < (NX+1)*(NY+1)
   * - the nodes have the coordinates  n_id = n_iy * (NX+1) + n_iy
   *
   * MESH Coordinate
   * - the transformation between mesh-coordinate and element / node coordinates is
   * -   x = _X0 + H * ix  /  y = _Y0 + H * iy
   *
   * VELOCITY:
   * - The vectors _VX, _VY store the Navier-Stokes velocity in each node
   *
   **/

  class Lattice
  {
  public:                                 //  protected:

    // Global Lattice Data
    double H;      // size of the lattice H^DIM
    int NX, NY;    // no of boxes
    double X0, Y0; // bounding box for lattice
    double X1, Y1;

    //
    std::vector<Element> elements;                // vector of elements
    std::vector<BoundaryLine> boundary_lines;     // vector of boundary lines
    std::vector<BoundaryCircle> boundary_circles; // vector of boundary circles


    ////////////////////////////////////////////////// Velocities
    std::vector<double> VX, VY; // velocities in each lattice node (NX+1)*(NY+1)

    ////////// Init...
    Lattice() {}

    void BasicInit(const Gascoigne::ParamFile &paramfile);

    // initializes the lattice from the gascoigne mesh
    // - elements are of size H^DIM
    // - the lattice covers the complete mesh

    // Sequence:
    //
    // 1) InitFromGascoigneMesh
    // 2) Lattice.AddBoundaryLine(...)
    //    Lattice.AddBoundaryCircle(...)
    // 3) InitBoundaries()
    void InitFromGascoigneMesh(const Gascoigne::GascoigneMesh2d *M, double H);
    void InitBoundaries(); //

    void AddBoundaryLine(const BoundaryLine &BL) { boundary_lines.push_back(BL); }
    void AddBoundaryCircle(const BoundaryCircle &BC) { boundary_circles.push_back(BC); }

    ////////// ID & COORDINATES
    // transformation between element id and element coordinates
    inline int e_id(int ix, int iy) const
    {
      assert(ix >= 0);
      assert(ix <= NX);
      assert(iy >= 0);
      assert(iy <= NY);
      return iy * NX + ix;
    }
    inline void e_id_to_e_coord(int id, int &ix, int &iy) const
    {
      iy = id / NX;
      assert(iy >= 0);
      assert(iy < NY);
      ix = id % NX;
      assert(ix >= 0);
    }
    int e_id_by_mesh_coord(double x, double y) const
    {
      assert(x >= X0);
      assert(x <= X1);
      assert(y >= Y0);
      assert(y <= Y1);
      int ix = (x - X0) / H;
      int iy = (y - Y0) / H;
      return e_id(ix, iy);
    }
    void mesh_coord_by_e_coord(double &x, double &y, int ix, int iy) const
    {
      assert(ix >= 0);
      assert(ix <= NX);
      assert(iy >= 0);
      assert(iy <= NY);
      x = X0 + H * ix;
      y = Y0 + H * iy;
    }
    void mesh_coord_by_e_id(double &x, double &y, int i) const
    {
      int ix, iy;
      e_id_to_e_coord(i, ix, iy);
      assert(ix >= 0);
      assert(ix <= NX);
      assert(iy >= 0);
      assert(iy <= NY);
      x = X0 + H * ix;
      y = Y0 + H * iy;
    }

    ////////// Elements
    // return element at lattice coordinate
    inline Element &element(int ix, int iy)
    {
      return elements[e_id(ix, iy)];
    }
    inline const Element &element(int ix, int iy) const
    {
      return elements[e_id(ix, iy)];
    }

    ////////////////////////////////////////////////// Particle
    // number of particles in the lattice
    int nparticles() const
    {
      int np = 0;
      for (const auto &el : elements)
        np += el.nparticles();
      return np;
    }
    // remove all particles with negative size
    void clear_deleted_particles();

    //////////// I/O
    void write_vtk(const std::string &name, int i);
    void write_lattice_vtk(const std::string &name, int i);
    void print_particles(std::ofstream &OUT) const
    {
      for (const auto &el : elements)
        el.print_particles(OUT);
      OUT << std::endl;
    }

    // Check the Particle system
    bool sanity_check()
    {
      for (const auto &el : elements)
        for (const auto &p : el._P)
          if (p.radius > 0) // don't check deleted particles
          {
            if (p.x < el.x0())
            {
              std::cerr << "particle left of element." << std::endl;
              return false;
            }
            if (p.y < el.y0())
            {
              std::cerr << "particle below element." << std::endl;
              return false;
            }
            if (p.x > el.x1())
            {
              std::cerr << "particle right of element." << std::endl;
              return false;
            }
            if (p.y > el.y1())
            {
              std::cerr << "particle above element." << std::endl;
              return false;
            }
    
          }
      return true;
    }

    // maxtimestep, H/|vmax|
    double dtmax()
    {
      double vmax = 0;
      for (const auto &el : elements)
        vmax = std::max(vmax, el.vmax());
      return H / (1.e-10 + vmax);
    }

    void delete_outside_particles();
    void delete_overlapping_particles();

    // initialization of random particles. something between [0,n) in every element
    // velocity in [-vmax,+vmax]
    void InitRandom(int n, double minradius, double maxradius, double density)
    {
      abort();
#pragma omp parallel for
      for (int i = 0; i < elements.size(); ++i)
        elements[i].InitRandom(n, minradius, maxradius, density, boundary_lines);
      // clean up list be removing all particles outside the domain
      delete_outside_particles();
      clear_deleted_particles();
    }

    // add new particles along first column
    void seed(int n, double minradius, double maxradius, double minheight, double maxheight)
    {
      for (int ix = 0; ix < NX; ++ix)

        for (int iy = 0; iy < NY; ++iy)
        {
          //// to get angle of incidence
          // std::array<double, 4> e_vx,e_vy; // velocity in element
          // transform element index to lattice node index
          int lni = iy * (NX + 1) + ix; // lattice node index...
          double vx = 0.25 * (VX[lni] + VX[lni + 1] + VX[lni + (NX + 1)] + VX[lni + (NX + 1) + 1]);
          double vy = 0.25 * (VY[lni] + VY[lni + 1] + VY[lni + (NX + 1)] + VY[lni + (NX + 1) + 1]);
          elements[iy * NX + ix].AddRandom(vx, vy,
                                           n,
					   minradius, maxradius,
					   minheight, maxheight, boundary_lines);
        }
      delete_outside_particles();
      delete_overlapping_particles();
      clear_deleted_particles();
    }

    // add a single particle
    void AddParticle(double x, double y, double vx, double vy, 
		     double height, double radius)
    {
      Particle p(x, y, vx, vy, height, radius);
      int e_id = e_id_by_mesh_coord(x, y);
      elements[e_id].AddParticle(p);
    }

    ////////////////////////////////////////////////// Motion of Particles
    // move all the particles by the given time-step. No control
    inline void advect(double dt)
    {
      ///////////////////////////////////// learning coefficients
      // once for creation of particles...
      Gascoigne::CompVector<double> parameters(6, nparticles(), 0);

#pragma omp parallel
      {
        std::array<double, 4> e_vx, e_vy; // velocity in element
     
#pragma omp for
        for (int i = 0; i < elements.size(); ++i)
        {
          // transform element index to lattice node index
          int ix = i % NX;
          int iy = i / NX;
          int lni = iy * (NX + 1) + ix;
         
          e_vx[0] = VX[lni];
          e_vy[0] = VY[lni];
          e_vx[1] = VX[lni + 1];
          e_vy[1] = VY[lni + 1];
          e_vx[2] = VX[lni + (NX + 1)];
          e_vy[2] = VY[lni + (NX + 1)];
          e_vx[3] = VX[lni + (NX + 1) + 1];
          e_vy[3] = VY[lni + (NX + 1) + 1];
          
          elements[i].advect(e_vx, e_vy, dt);
        }
      }
    }

    inline void collide_boundary()
    {
      
      for (int i = 0; i < elements.size(); ++i)
        elements[i].collide_boundary(boundary_lines, boundary_circles);
    }
    inline void collide_particles(double dt)
    {
        
#pragma omp parallel for
      for (int iy = 0; iy < NY; ++iy)
        for (int ix = 0; ix < NX; ++ix)
        {
          int eid = e_id(ix, iy);
         
          
          elements[eid].collide_particles(dt); // cell no. 0 (center)

          // oben rechts
          if( (ix < NX-1) && (iy < NY-1) ){
            elements[eid].collide_particles(elements[eid + NX + 1], dt); 
          }

          // oben links
          if( (ix > 0) && (iy < NY-1) ){
            elements[eid].collide_particles(elements[eid + NX - 1], dt); 
          }

          if (ix < NX - 1){
            elements[eid].collide_particles(elements[eid + 1], dt); //to the right
          }
        } 
#pragma omp parallel for
      for (int ix = 0; ix < NX; ++ix)
        for (int iy = 0; iy < NY; ++iy)
        {
          int eid = e_id(ix, iy);
          if (iy < NY - 1){
            elements[eid].collide_particles(elements[eid + NX], dt); //to the top
          }        
        }
    }

    // transfer to particles to neighbours if necessary
    inline void transfer()
    {
   
      // first horizontal:
#pragma omp parallel for
      for (int iy = 0; iy < NY; ++iy)
        for (int ix = 0; ix < NX; ++ix)
        {
         
          int eid = e_id(ix, iy);
          for (int p = 0; p < elements[eid].nparticles(); ++p)
          {
            
            if (elements[eid].particle(p).x < elements[eid].x0()) // left
            {
              
              if (ix > 0)
               { 
              elements[eid - 1].AddParticle(elements[eid].particle(p));
              elements[eid].particle(p).radius = -1; // mark for deletion
              
              } else {
                elements[eid].particle(p).x = elements[eid].x0();
              }

            }
            if (elements[eid].particle(p).x > elements[eid].x1()) // right
            {
              
              if (ix < NX - 1) {
                elements[eid + 1].AddParticle(elements[eid].particle(p));
            
              elements[eid].particle(p).radius = -1; // mark for deletion
              
              } else {
                elements[eid].particle(p).x = elements[eid].x1();
              }
              
            }
           
          }
        
        }
       
        // second vertical
#pragma omp parallel for
      for (int ix = 0; ix < NX; ++ix)
        for (int iy = 0; iy < NY; ++iy)
        {
          int eid = e_id(ix, iy);
          for (int p = 0; p < elements[eid].nparticles(); ++p)
          {
             
            if (elements[eid].particle(p).y < elements[eid].y0()) // down
            {
             
              if (iy > 0) {
                elements[eid - NX].AddParticle(elements[eid].particle(p));
              elements[eid].particle(p).radius = -1; // mark for deletion
              } else {
                elements[eid].particle(p).y = elements[eid].y0();
              }
            }
            if (elements[eid].particle(p).y > elements[eid].y1()) // up
              
            {
              if (iy < NY - 1) {
                elements[eid + NX].AddParticle(elements[eid].particle(p));
              
              elements[eid].particle(p).radius = -1; // mark for deletion
             
              } else {
                elements[eid].particle(p).y = elements[eid].y1();
              }
            }
          
          }
          
        }
      
    }
  };

}

/*----------------------------   lattice.h     ---------------------------*/
/* end of #ifndef __lattice_H */
#endif
/*----------------------------   lattice.h     ---------------------------*/
