/*----------------------------   particle.h     ---------------------------*/
/*      $Id:$                 */
#ifndef __particle_H
#define __particle_H
/*----------------------------   particle.h     ---------------------------*/

//#include "boundaryline.h"
#include <fstream>
#include <cmath>
#include <complex> // std::complex, l2_norm
#include <iostream>
#include <vector>
#include <cassert>

extern double TIME;

namespace GasParticle
{
  extern int NPARTICLES;

  
#define ICEDENSITY 900.0

  

  class Particle
  {
  public:
    // data
    int particle_id;

    double x, y;                  // position
    double vx, vy;                // translational velocity
    double marker;
    double collision1, collision2, collision3;
    

    // double omega, angle;          // angular velocity & angle between longest axis of the particle (in the velocity direction) and vertical axis of the domain
    double height;
    double radius;      
    double Cd, Cl, Ct, Cp; // force coefficients

    Particle(const Particle &p) : particle_id(p.particle_id),
                                  x(p.x), y(p.y),
                                  vx(p.vx), vy(p.vy),
                                  //omega(p.omega), angle(p.angle),
                                  height(p.height), radius(p.radius),
                                  marker(p.marker),
                                  collision1(p.collision1), collision2(p.collision2), collision3(p.collision3)
    {
    }

    Particle() : x(0), y(0), vx(0), vy(0),
		 //omega(0), angle(0),
                 height(0), radius(0), //{ assert(0); }
                 marker(0),
                 collision1(0), collision2(0), collision3(0) { assert(0); }
    Particle(double _x, double _y, double _vx, double _vy,
	     //double _omega, double _angle,
             double _height, double _radius) : x(_x), y(_y), vx(_vx), vy(_vy),
					       //omega(_omega), angle(_angle),
					       height(_height), radius(_radius)
    {
#pragma omp critical
      particle_id = NPARTICLES++;
    }

    //just l2_norm
    inline double l2_norm(std::vector<double> const &u) const
    {
      double accum = 0.;
      for (int i = 0; i < u.size(); ++i)
	{
	  assert(finite(u[i]));
	  accum += u[i] * u[i];
	}
      return sqrt(accum);
    }
    inline double l2_norm(double const &u) const
    {
      return std::abs(u);
    }

    inline double mass() const
    {
      // return M_PI * radius * radius * height * ICEDENSITY; // Sphere
      return 4* radius * radius * height * ICEDENSITY;
    }
    inline double area() const
    {
       //return 4 * radius * radius;
      return M_PI * radius * radius;
    }
    inline double volume() const
    {
      return area() * height;
    }
    inline double momentofinertia() const
    {
      std::cerr << "moment of inertia of a disc" << std::endl;
      abort();
    }
    inline std::vector<double> rel_vel(double ns_vx, double ns_vy) const // the particle velocity relative to the local fluid velocity
    {
      std::vector<double> temp(2);
      temp[0] = ns_vx - vx;
      temp[1] = ns_vy - vy;
      return temp;
    }
    inline double rel_rot(double ns_omega) const // the particle relative rotation
    {
      abort();
      
      // // should be curl(ns_v - v_particle) - omega_particle
      // // in element.h curl_v is used as an argument for ns_omega when function advect is called
      // return ns_omega - omega;
    }

    // advection
    inline void advect(double ns_vx, double ns_vy, double ns_omega, double dt)
    {
      if (radius < 0)
        return; // skip deleted particle with negative radius

 
      
      double Cd = 1.0; // modeling dynamics and thermodynamics of iceberg drag 0.9

      vx = ns_vx;
      vy = ns_vy;
      
      assert(finite(vx));
      assert(finite(vy));



      // update location
      x += dt * vx;
      y += dt * vy;
    
 
      assert(finite(x));
      assert(finite(y));

    
    }

    // necessary element.h for calculating maximum velocity
    inline double vnorm2() const { return vx * vx + vy * vy; }
    inline double vnorm() const { return sqrt(vnorm2()); }

    // check distance between two particles measured from surface
    // negative distance is collision
    inline double outer_distance(const Particle &p) const
    {
       
      return sqrt((p.x - x) * (p.x - x) + (p.y - y) * (p.y - y)) - (radius + p.radius);
    }

    // I/O
    void print(std::ofstream &OUT) const
    {
      if (radius < 0)
        return;
      OUT << x << " " << y << std::endl;
    }
  };
  std::ostream &operator<<(std::ostream &out, const Particle &p);
}

/*----------------------------   particle.h     ---------------------------*/
/* end of #ifndef __particle_H */
#endif
/*----------------------------   particle.h     ---------------------------*/
